/*
 * Copyright (c) 2007 Kathryn Huxtable
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: UserAttributesImpl.java 1087 2007-04-10 15:39:08Z kathryn5 $
 */
package org.kathrynhuxtable.middleware.shibshim.filter;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Vector;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.apache.log4j.Logger;
import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.xml.sax.SAXException;

/**
 * Attributes object for the Shibboleth shim filter. Hold the collection of
 * attributes and provide parsing method.
 */
public class UserAttributesImpl extends HashMap implements UserAttributes {
    /**
     * Logger for error/information/debugging.
     */
    private static Logger       log                 = Logger.getLogger(UserAttributesImpl.class.getName());

    /**
     * Initial capacity for hashes.
     */
    private static final int    INITIAL_CAPACITY    = 5;

    /**
     * Text of the DTD for the XML assertion. I'm not convinced that we need to
     * verify this.
     */
    private static final String DTD_TEXT            = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<!DOCTYPE  Assertion [\n"
                                                            + "<!ELEMENT  Assertion (AttributeStatement)>\n<!ATTLIST  Assertion\n"
                                                            + "  IssueInstant CDATA #REQUIRED\n  AuthenticationInstant CDATA #REQUIRED\n"
                                                            + "  Issuer CDATA #REQUIRED\n  Handle CDATA #REQUIRED>\n"
                                                            + "<!ELEMENT  AttributeStatement (Attribute*)>\n<!ELEMENT  Attribute (AttributeValue+)>\n"
                                                            + "<!ATTLIST  Attribute\n  AttributeName CDATA #REQUIRED>\n"
                                                            + "<!ELEMENT  AttributeValue (#PCDATA | AttributeStatement)*>\n]>\n";

    // Constants used while parsing the user information retrieved in XML
    // format.

    /**
     * Name of the &lt;Attribute&gt; element.
     */
    private static final String ATTRIBUTE_TAG       = "Attribute";

    /**
     * Name of the &lt;AttributeName&gt; attribute.
     */
    private static final String ATTRIBUTE_NAME_TAG  = "AttributeName";

    /**
     * Name of the &lt;AttributeValue&gt; element.
     */
    private static final String ATTRIBUTE_VALUE_TAG = "AttributeValue";

    /**
     * Name of the &lt;Assertion&gt; element.
     */
    private static final String ASSERTION_TAG       = "Assertion";

    /**
     * Name of the &lt;AttributeStatement&gt; element.
     */
    private static final String ATTR_STMT_TAG       = "AttributeStatement";

    /**
     * Indicates whether object has user attributes, as distinct from
     * configuration attributes.
     */
    private boolean             hasAttributes;

    /**
     * Map mapping incoming Shibboleth attributes to application attributes. Any
     * attribute not in the key set will be ignored. Case is ignored for keys
     * but not for values.
     */
    private Map                 attributeMap;

    /**
     * Make new object. Sets hasAttributes false.
     * 
     * @param attributeMap
     *            the map mapping incoming Shibboleth attribute names to
     *            internal attribute names. Any attributes not listed here will
     *            be ignored.
     */
    public UserAttributesImpl(Map attributeMap) {
        super();
        hasAttributes = false;
        this.attributeMap = attributeMap;
    }

    /**
     * {@inheritDoc}
     */
    public boolean isMappedAttribute(String name) {
        return attributeMap.containsValue(name.toLowerCase());
    }

    /**
     * {@inheritDoc}
     */
    public String getHeader(String name) {
        ArrayList list = null;
        name = name.toLowerCase();
        if (name.startsWith("config-")) {
            return (String) get(name);
        }
        list = (ArrayList) get(name);
        log.debug("Got list = " + list + ", size = " + (list == null ? "?" : (new Integer(list.size())).toString()));
        Object res = list == null || list.size() == 0 ? null : list.get(0);
        if (res instanceof HashMap) {
            // This is a kuorganizationalrole element.
            HashMap subMap = (HashMap) res;
            Iterator itr = (subMap.keySet()).iterator();
            String subTable = null;
            subTable = "<Assertion>";
            while (itr.hasNext()) {
                String subName = (String) itr.next();
                subTable += "<Attribute name=\"" + subName + "\">" + subMap.get(subName) + "</Attribute>";
            }
            subTable += "</Assertion>";
            return subTable;
        }
        return (String) res;
    }

    /**
     * {@inheritDoc}
     */
    public Enumeration getHeaders(String name) {
        ArrayList list = null;

        name = name.toLowerCase();

        if (name.startsWith("config-")) {
            Vector res = new Vector();
            res.add(get(name));
            return res.elements();
        } else {
            list = (ArrayList) get(name);
            if (list == null) {
                return (new Vector()).elements();
            }
        }

        return (new Vector(list)).elements();
    }

    /**
     * {@inheritDoc}
     */
    public Enumeration getHeaderNames() {
        Vector v = new Vector();
        Iterator keys = keySet().iterator();
        while (keys.hasNext()) {
            String key = (String) keys.next();
            v.add(key);
        }
        return v.elements();
    }

    /**
     * Return whether we have user attributes or just configuration attributes.
     * 
     * @return true if we have user attributes
     */
    public boolean hasAttributes() {
        return hasAttributes;
    }

    /**
     * Clear the user attributes, leaving the configuration attributes alone.
     */
    public void clearUserInfo() {
        Iterator keys = this.keySet().iterator();
        Vector remMap = new Vector();
        int track = 0;
        while (keys.hasNext()) {
            String key = (String) keys.next();
            if (!key.startsWith("config-")) {
                remMap.add(key);
            }
        }
        while (track < remMap.size()) {
            this.remove(remMap.get(track));
            track++;
        }
        hasAttributes = false;
    }

    /**
     * Parse the user info string from the Shibboleth Shim server into a hash
     * map object.
     * 
     * @param userInfo
     *            String containing the user info in XML format.
     * @throws UserInfoException
     *             if an error occurs.
     */
    public void parseUserInfo(String userInfo) throws UserInfoException {
        // Parse the XML string and obtain a DOM document.
        Document doc = parseXmlFile(userInfo);

        // Top element must be ASSERTION_TAG.
        Element top = doc.getDocumentElement();
        if (!top.getTagName().equals(ASSERTION_TAG)) {
            return;
        }

        // Retrieve the Assertion attributes.
        NamedNodeMap assertionAttrs = top.getAttributes();
        for (int i = 0; i < assertionAttrs.getLength(); i++) {
            Attr attr = (Attr) assertionAttrs.item(i);
            ArrayList assertionList = new ArrayList(INITIAL_CAPACITY);
            assertionList.add(attr.getNodeValue());
            this.put("config-" + attr.getNodeName().toLowerCase(), assertionList);
        }

        // Assertion must contain one and only one ATTR_STMT_TAG element.
        Node attributeStatement = top.getFirstChild();
        if (attributeStatement == null || !attributeStatement.getNodeName().equals(ATTR_STMT_TAG) || attributeStatement.getNextSibling() != null) {
            return;
        }

        // Loop over ATTR_STMT_TAG elements.
        for (Node attribute = attributeStatement.getFirstChild(); attribute != null; attribute = attribute.getNextSibling()) {
            String nodeName = attribute.getNodeName();
            // Skip non-Attribute elements. There shouldn't be any...
            if (nodeName == null || !nodeName.equals(ATTRIBUTE_TAG)) {
                continue;
            }
            // Retrieve the attribute name. Skip if not an attributeMap key.
            String attrName = ((Element) attribute).getAttribute(ATTRIBUTE_NAME_TAG);
            if (!attributeMap.containsKey(attrName.toLowerCase())) {
                log.debug("Skipping incoming attribute \"" + attrName + "\"");
                continue;
            }
            // For every "Attribute" element get the attribute name and its
            // value.
            ArrayList attrList = new ArrayList(INITIAL_CAPACITY);
            for (Node value = attribute.getFirstChild(); value != null; value = value.getNextSibling()) {
                String valueNodeName = value.getNodeName();
                // Skip anything outside a "Value" tag. There shouldn't be
                // anything...
                if (valueNodeName == null || !valueNodeName.equals(ATTRIBUTE_VALUE_TAG)) {
                    continue;
                }
                StringBuffer attrValue = new StringBuffer();
                for (Node valueNode = value.getFirstChild(); valueNode != null; valueNode = valueNode.getNextSibling()) {
                    if (valueNode.getNodeName() == null || !valueNode.getNodeName().equals(ATTR_STMT_TAG)) {
                        attrValue.append(valueNode.getNodeValue());
                    } else {
                        // Parse an embedded <AttributeStatement> as a value
                        // element.
                        // This involves reconstructing a String
                        // representation of the XML.
                        attrValue.append(encodeAttributeStatement(valueNode));
                    }
                }
                if (attrValue.length() > 0) {
                    hasAttributes = true;
                    attrList.add(attrValue.toString());
                }
            }
            if (!attrList.isEmpty()) {
                log.debug("Saving attribute \"" + attributeMap.get(attrName.toLowerCase()) + "\" with value " + attrList);
                this.put(attributeMap.get(attrName.toLowerCase()), attrList);
            }
        }
    }

    /**
     * Parses an XML file and returns a DOM document.
     * 
     * @param userInfo
     *            The user information in XML Format.
     * 
     * @return Document A DOM Document.
     * 
     * @throws UserInfoException
     *             if an error occurs.
     */
    private static Document parseXmlFile(String userInfo) throws UserInfoException {
        try {
            userInfo = DTD_TEXT + userInfo;

            // Create a builder factory and set it to validate.
            DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
            factory.setValidating(true);
            log.debug("User Info = \"" + userInfo + "\"");

            // Create the builder and parse the file
            ByteArrayInputStream inputStream = new ByteArrayInputStream(userInfo.getBytes());
            return factory.newDocumentBuilder().parse(inputStream);
        } catch (SAXException e) {
            log.debug("ShibShimFilter::parseXmlFile:Received a SAXException::" + e.getMessage());
            throw new UserInfoException("ShibShimFilter::parseXmlFile:Received a SAXException::" + e.getMessage());
        } catch (ParserConfigurationException e) {
            log.debug("ShibShimFilter::parseXmlFile:Received a ParserConfigurationException::" + e.getMessage());
            throw new UserInfoException("ShibShimFilter::parseXmlFile:Received a ParserConfigurationException::" + e.getMessage());
        } catch (IOException e) {
            log.debug("ShibShimFilter::parseXmlFile:Received a IOException::" + e.getMessage());
            throw new UserInfoException("ShibShimFilter::parseXmlFile:Received an IOException::" + e.getMessage());
        }
    }

    /**
     * Encode the attributes in a sub-object into an XML document.
     * 
     * @param attributeStatement
     *            The attributeStatement node parsed from the XML.
     * @return the encoded attribute statement.
     */
    private StringBuffer encodeAttributeStatement(Node attributeStatement) {
        StringBuffer result = new StringBuffer();
        result.append("<" + ATTR_STMT_TAG + ">");

        // Loop over ATTR_STMT_TAG elements.
        for (Node attribute = attributeStatement.getFirstChild(); attribute != null; attribute = attribute.getNextSibling()) {
            String nodeName = attribute.getNodeName();
            // Skip non-Attribute elements. There shouldn't be any...
            if (nodeName == null || !nodeName.equals(ATTRIBUTE_TAG)) {
                continue;
            }
            // Retrieve the attribute name.
            String attrName = ((Element) attribute).getAttribute(ATTRIBUTE_NAME_TAG);
            // For every "Attribute" element get the attribute name and its
            // value.
            result.append("<" + ATTRIBUTE_TAG + " " + ATTRIBUTE_NAME_TAG + "=\"" + attrName + "\">");
            for (Node value = attribute.getFirstChild(); value != null; value = value.getNextSibling()) {
                String valueNodeName = value.getNodeName();
                // Skip anything outside a "Value" tag. There shouldn't be
                // anything...
                if (valueNodeName == null || !valueNodeName.equals(ATTRIBUTE_VALUE_TAG)) {
                    continue;
                }
                StringBuffer attrValue = new StringBuffer();
                for (Node valueNode = value.getFirstChild(); valueNode != null; valueNode = valueNode.getNextSibling()) {
                    if (valueNode.getNodeName() == null || !valueNode.getNodeName().equals(ATTR_STMT_TAG)) {
                        attrValue.append(valueNode.getNodeValue());
                    } else {
                        // Parse an embedded <AttributeStatement> as a value
                        // element.
                        // This involves reconstructing a String
                        // representation of the XML.
                        attrValue.append(encodeAttributeStatement(valueNode));
                    }
                }
                if (attrValue.length() > 0) {
                    result.append("<" + ATTRIBUTE_VALUE_TAG + ">");
                    result.append(attrValue.toString());
                    result.append("</" + ATTRIBUTE_VALUE_TAG + ">");
                }
            }
            result.append("</" + ATTRIBUTE_TAG + ">");
        }

        result.append("</" + ATTR_STMT_TAG + ">");
        return result;
    }
}
